#pragma once
#include <string>
#include <map>
#include "llvm/IR/LLVMContext.h"
#include "llvm/IR/Type.h"
#include "llvm/IR/DerivedTypes.h"

namespace type_info {
    enum class TypeValue {
        CHARACTER,
        SHORT,
        INTEGER,
        LONG,
        LONGLONG,
        FLOAT,
        DOUBLE,
        BOOLEAN,
        CUSTOM,
        ERROR,
        UNKNOWN,
        VOID,
    };

    enum class TypeQualityVal {
        VOLATILE,
        CONST,
        VOLATILE_CONST,
        NONE,
        UNKNOWN,
        ERROR,
    };

    enum class StorageSpecVal {
        TYPEDEF,
        EXTERN,
        STATIC,
        REGISTER,
        QUANTUM,
        AUTO,
        UNKNOWN,
        ERROR,
    };

    class Type;

    bool operator==(const Type& lhs, const Type& rhs);
    bool operator!=(const Type& lhs, const Type& rhs);

    class Type {
        public:
            // Default constructor
            Type();

            // Copy Constructor
            Type(const Type& other);

            // Constructor for scalar types
            Type(TypeValue typeVal, bool signedVal = true,
                 TypeQualityVal qualityVal = TypeQualityVal::NONE,
                 StorageSpecVal storageVal = StorageSpecVal::AUTO);

            // Constructor for arrays
            Type(std::shared_ptr<Type> typeVal,
                int arraySize,
                TypeQualityVal qualityVal = TypeQualityVal::NONE,
                StorageSpecVal storageVal = StorageSpecVal::AUTO);

            // Constructor for functions
            Type(std::shared_ptr<Type> typeVal,
                 std::vector<Type> argTypes,
                 bool varArg = false,
                 TypeQualityVal qualityVal = TypeQualityVal::NONE,
                 StorageSpecVal storageVal = StorageSpecVal::EXTERN);

            // Constructor for pointers
            Type(std::shared_ptr<Type> typeVal,
                 TypeQualityVal qualityVal = TypeQualityVal::NONE,
                 StorageSpecVal storageVal = StorageSpecVal::AUTO);

            // Return string for value
            std::string get_value_string() const;
            std::string get_quality_string() const;
            std::string get_storage_string() const;

            // Get base values for complex types
            std::pair<Type *, bool> get_base_type();
            TypeValue get_base_type_value() const;
            TypeQualityVal get_base_type_quality() const;
            StorageSpecVal get_base_type_storage() const;

            void set_base_type(Type newBase);

            // Methods for generating LLVM type information
            llvm::Type * codegen_type(llvm::LLVMContext& context);
            // llvm::GlobalValue::LinkageType * codegen_storage(llvm::Context context);

            std::pair<std::string, int> quantum_codegen_type();

            // Methods for generating basic size information for basic types
            static int getTypeBitSize(Type type);

            bool typeValCompare(Type other);

            // Return string for complete type info (includes array and function info)
            operator std::string() const;

            // Comparison operators
            friend bool type_info::operator==(const Type& lhs, const Type& rhs);
            friend bool type_info::operator!=(const Type& lhs, const Type& rhs);

            // Generic type info;
            // Fully applicalbe if a scalar type, but only partially otherwise
            TypeValue value;
            bool isSigned;
            TypeQualityVal quality;
            StorageSpecVal storage;

            // Recursive type info for more complex types
            std::shared_ptr<Type> subType;

            // Scalar-specific type info
            bool isScalar;

            // Array-specific type info
            int size;
            bool isArray;

            // Function-related type info
            bool isFunction;
            std::vector<Type> functionArgTypes;
            bool isVarArg;

            // Pointer-related type information
            bool isPointer;

            // TODO: handle structs/unions
    };

    enum class KindValue {
        LABEL,
        UNINITVAR,
        INITVAR,
        VALUE,
        FUNC,
        FUNCARG,
        OTHER
    };

    class Kind {
        public:
            Kind();
            Kind(KindValue kind);

            // Return string for complete type info (includes array and function info)
            operator std::string() const;

            KindValue value;
    };



}
