#include <iostream>
#include <iomanip>
#include <algorithm>
#include <iterator>
#include <string_view>
#include <cmath>
#include <climits>
#include <exception>
#include <stdexcept>
#include <cstdlib>
#include "symbol_table/symbol_table.hpp"
#include "ast/ast.hpp"
#include "quant_c_parse.hpp"
#include "llvm/ADT/APInt.h"

using namespace quant_c_ast;
using namespace symbol_table;
using namespace type_info;
using namespace quant_c_compiler;

// abstract superclass; only subclasses are actually created
DeclNode::DeclNode(ASTNodeKind nodeKind) : ASTNode::ASTNode(nodeKind) {}
DeclNode::DeclNode(ASTNodeKind nodeKind, int line, int col) : ASTNode::ASTNode(nodeKind, line, col) {}
bool DeclNode::classof(const ASTNode * node) {
    auto nodeKind = node->getNodeKind();
    return nodeKind >= ASTNodeKind::AST_DeclNode &&
           nodeKind < ASTNodeKind::AST_NullTypeNode;
} // classof
// class DeclNode

DeclarationNode::DeclarationNode(
        std::vector<std::shared_ptr<TypeNode>> declSpec,
        std::vector<std::shared_ptr<InitDeclaratorNode>> initList,
        int line, int col) :
DeclNode(ASTNodeKind::AST_DeclarationNode, line, col) {
    this->declSpec = declSpec;
    this->initList = initList;
}
void DeclarationNode::unparse(int indent) {
    std::cout << std::left << std::setw(4) << linenum << ":";
    genIndent(indent);
    for (auto&type : declSpec) {
        type->unparse();
        std::cout << " ";
    }
    std::cout << " ";
    for (auto& decl: initList) {
        decl->unparse();
        std::cout << " , ";
    }

    // Back up to delete the last comma if arguments are printed
    if (initList.size() > 0) {
        std::cout << "\b\b\b";
    }
    std::cout << " ;\n";
}
type_info::Type DeclarationNode::generate_type() {
    Type argTypeSpecs = Type();
    argTypeSpecs.isScalar = true;
    bool set_signedness = false;
    bool set_storage = false;
    bool set_quality = false;

    for (auto& spec : declSpec) {
        auto type_node = spec.get();
        if (llvm::isa<TypeSpecNode>(type_node)) {
            if (llvm::isa<LongTypeNode>(type_node)) {
                if (argTypeSpecs.value == TypeValue::UNKNOWN ||
                    argTypeSpecs.value == TypeValue::INTEGER) {
                    argTypeSpecs.value = TypeValue::LONG;
                }
                else if (argTypeSpecs.value == TypeValue::LONG) {
                    argTypeSpecs = TypeValue::LONGLONG;
                }
                else {
                    mustBe(false, error() + "both 'long' and '" +
                           argTypeSpecs.get_value_string() +
                           "' in declaration specifiers");
                    return Type(TypeValue::ERROR);
                }
            }
            else if (llvm::isa<ShortTypeNode>(type_node)) {
                if (argTypeSpecs.value == TypeValue::UNKNOWN ||
                    argTypeSpecs.value == TypeValue::INTEGER) {
                    argTypeSpecs.value = TypeValue::SHORT;
                }
                else if (argTypeSpecs.value == TypeValue::SHORT) {
                    mustBe(false, error() + "duplicate 'short'");
                    return Type(TypeValue::ERROR);
                }
                else {
                    mustBe(false, error() + "both 'short' and '" +
                           argTypeSpecs.get_value_string() +
                           "' in declaration specifiers");
                    return Type(TypeValue::ERROR);
                }
            }
            else if (llvm::isa<SignedTypeNode>(type_node)) {
                if (set_signedness) {
                    if (argTypeSpecs.isSigned) {
                        mustBe(false, error() + "duplicate 'signed'");
                    }
                    else {
                        mustBe(false, error() + "both 'signed' and 'unsigned' in declaration specifiers");
                    }
                    return Type(TypeValue::ERROR);
                } else {
                    argTypeSpecs.isSigned = true;
                    set_signedness = true;
                }
            }
            else if (llvm::isa<UnsignedTypeNode>(type_node)) {
                if (set_signedness) {
                    if (!argTypeSpecs.isSigned) {
                        mustBe(false, error() + "duplicate 'unsigned'");
                    }
                    else {
                        mustBe(false, error() + "both 'signed' and 'unsigned' in declaration specifiers");
                    }
                    return Type(TypeValue::ERROR);
                } else {
                    argTypeSpecs.isSigned = false;
                    set_signedness = true;
                }
            }
            else if (argTypeSpecs.value == TypeValue::UNKNOWN) {
                argTypeSpecs.value = spec->type.value;
            }
            else {
                mustBe(false, error() +
                       "two or more data types in declaration specifiers");
                return Type(TypeValue::ERROR);
            }
        }
        else if (llvm::isa<StorageClassSpecNode>(type_node)) {
            if (set_storage) {
                mustBe(false, error() + "multiple storage classes in declaration specifier");
            }
            else if (llvm::isa<AutoNode>(type_node)) {
                if(isGlobalDec) {
                    mustBe(false, error() + "illegal storage class on file-scoped variable/function");
                }
                else {
                    argTypeSpecs.storage = StorageSpecVal::AUTO;
                    set_storage = true;
                }
            }
            else if (llvm::isa<TypeDefNode>(type_node)) {
                argTypeSpecs.storage = StorageSpecVal::TYPEDEF;
                set_storage = true;
            }
            else if (llvm::isa<ExternNode>(type_node)) {
                argTypeSpecs.storage = StorageSpecVal::EXTERN;
                set_storage = true;
            }
            else if (llvm::isa<StaticNode>(type_node)) {
                argTypeSpecs.storage = StorageSpecVal::STATIC;
                set_storage = true;
            }

            else if (llvm::isa<RegisterNode>(type_node)) {
                if(isGlobalDec) {
                    mustBe(false, error() + "illegal storage class on file-scoped variable/function");
                }
                else {
                    argTypeSpecs.storage = StorageSpecVal::REGISTER;
                    set_storage = true;
                }
            }
            else { // llvm::isa<QuantumNode>(type_node)
                argTypeSpecs.storage = StorageSpecVal::QUANTUM;
                set_storage = true;
            }
        }
        else if (llvm::isa<TypeQualifierNode>(type_node)) {
            set_quality = true;
            if (argTypeSpecs.quality == TypeQualityVal::VOLATILE_CONST) {
                continue; // Any additional qualifiers are ignored
            }
            else if (llvm::isa<VolatileNode>(type_node)) {
                argTypeSpecs.quality =
                    (argTypeSpecs.quality == TypeQualityVal::CONST) ?
                    TypeQualityVal::VOLATILE_CONST :
                    TypeQualityVal::VOLATILE;

            } else { // llvm::isa<ConstNode>(spec)
                argTypeSpecs.quality =
                    (argTypeSpecs.quality == TypeQualityVal::VOLATILE) ?
                    TypeQualityVal::VOLATILE_CONST :
                    TypeQualityVal::CONST;
            }
        }
        else {
            mustBe(false, error() + "Unimplemented type used");
            return Type(TypeValue::ERROR);
        }
    }

    // Establish defaults if left implicit in the type specs
    if (argTypeSpecs.value == TypeValue::UNKNOWN) {
        argTypeSpecs.value = TypeValue::INTEGER;
    }
    if (!set_signedness) {
        argTypeSpecs.isSigned = true;
    }
    if (!set_storage) {
        if (isGlobalDec) {
            argTypeSpecs.storage = StorageSpecVal::EXTERN;
        }
        else {
            argTypeSpecs.storage = StorageSpecVal::AUTO;
        }
    }
    if (!set_quality) {
        argTypeSpecs.quality = TypeQualityVal::NONE;
    }
    return argTypeSpecs;
}

void DeclarationNode::checkSemantics() {
    Type decl_type = generate_type();

    // Perform the semantic checks on the initList
    // and check
    for (auto& initDecl : initList) {
        initDecl->quality = decl_type.quality;
        initDecl->checkSemantics();

        // Attempt to lookup decl in symbol table
        // and perform some checks based on
        // the declared type and kind of declarator
        // TODO: implement checks for other kinds of declarations
        try {

            SymbolInfo& id = semanticTable.localLookup(initDecl->id_name);

            //auto declPtr = initDecl->decl->direc.get();
            if (id.kind.value == KindValue::INITVAR ||
                id.kind.value == KindValue::UNINITVAR) {
                // Update the symbol's type info to reflect it's type
                // In particular, the base type
                // (ex. array element type)
                if (id.type.isScalar) {
                    id.type = decl_type;
                }
                else {
                    /*
                    auto [baseType, hasSubType] =
                                                    id.type.get_base_type();
                    if (!hasSubType) {
                        baseType->subType = std::make_shared<Type>();
                        *(baseType->subType) = decl_type;
                    }
                    else {
                        baseType = std::make_shared<Type>();
                        *baseType = decl_type;
                    }
                    */
                    id.type.set_base_type(decl_type);
                }

                // TODO: Check that the initializer expression (if present)
                // is convertable to the declared type
            }
            else if(id.kind.value == KindValue::FUNC) {
                // Move storage information into the function type
                // itself rather than return type
                // and update type
                id.type.storage = decl_type.storage;
                decl_type.storage = StorageSpecVal::AUTO;


                // Update the symbol's type info to reflect it's return type
                // TODO: look into destructuring
                /*
                auto [baseType, hadValidSubType] = id.type.get_base_type();
                if (!hadValidSubType) {
                    // Means that the type had a nullptr subtype,
                    // even though it shouldn't; report error
                    // and create new copy of type with pointer
                    // mustBe(false, error() + "Unexpected failure while setting type information for declaration; this must be a compiler bug.");
                    baseType->subType = std::make_shared<Type>();
                    *(baseType->subType) = decl_type;
                }
                else {
                    // Directly set the base type to be the declared type
                    *(baseType) = decl_type;
                }
                */
                id.type.set_base_type(decl_type);


                /*
                auto sharedDeclType = std::make_shared<Type>(decl_type);
                id.type = Type(sharedDeclType, id.type.functionArgTypes,
                               decl_type.isSigned,
                               id.type.isVarArg,
                               decl_type.quality, decl_type.storage);
                               */
            }

            // Save symbol info for code generation
            id_info.push_back(id);
        }
        catch (FailedLookupException &lookupEx) {
            // Shouldn't occur
            mustBe(false, error() + "Unexpected failure while checking program; this must be a compiler bug.");
        }
    }
}

llvm::Value* DeclarationNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    // Generate code for each declaration and add type information
    int index = 0;
    for (auto& initDecl : initList) {
        Type declType = id_info[index].type;
        llvm::Type* declGenType = declType.codegen_type(LLVMContext);
        auto declName = initDecl->decl->id_name;
        // Have to pass type info; form some reason, trying to change the type
        // of a AllocaInst via `setAllocatedType` just does nothing
        llvm::Value* declVal = initDecl->codegen(nullptr, declGenType);

        // TODO: verify validity
        if(llvm::isa<llvm::Function>(declVal)) {

            llvm::Function* basicFuncDecl =
                llvm::cast<llvm::Function>(declVal);
            /*
            llvm::ArrayRef<llvm::Type *> argTypes = basicFuncDecl->
                                                     getFunctionType()->
                                                     params();
            llvm::FunctionType* funcType =
                llvm::FunctionType::get(declGenType,
                                        argTypes,
                                        basicFuncDecl->isVarArg());
                                        */
            // Create full function declaration
            // The type info created during semantics checking should be
            // sufficient
            auto linkage = (declType.storage ==
                            StorageSpecVal::EXTERN) ?
                           llvm::Function::ExternalLinkage :
                           llvm::Function::PrivateLinkage;
            llvm::FunctionType* funcType =
                llvm::cast<llvm::FunctionType>(declGenType);
            llvm::Function* functionCode =
                llvm::Function::Create(funcType,
                                       linkage,
                                       declName,
                                       LLVMModule.get());

            // Add argument names to final function declaration
            // from original function declaration
            llvm::Function::arg_iterator declArgIter =
                basicFuncDecl->arg_begin();
            for (llvm::Function::arg_iterator newArgIter =
                 functionCode->arg_begin();
                 newArgIter != functionCode->arg_end();
                 newArgIter++) {
                    auto argName = declArgIter->getName();
                    newArgIter->setName(argName);
                    declArgIter++;
            }

            // Verify generated code is a valid function declaration
            // and return
            llvm::verifyFunction(*functionCode);
            // return functionCode;
        }
        else if(llvm::isa<llvm::AllocaInst>(declVal)) {
            llvm::AllocaInst* varInstruction =
                                    llvm::cast<llvm::AllocaInst>(declVal);

            // Put constructed instruction into code
            llvm::Value* variableValue = nullptr;
            if (codegenTable.scopeDepth() == 1) {
                // Create as global variable
                variableValue = LLVMModule->getOrInsertGlobal(
                                                varInstruction->getName(),
                                                varInstruction->getType());

            } else {
                variableValue =
                        Builder.CreateAlloca(varInstruction->getAllocatedType(),
                                             varInstruction->getArraySize(),
                                             declName);
            }

            // Generate initialization code here (instead of in InitDecl)
            if(initDecl->hasInitalizer) {
                auto initializerCode = initDecl->initializer->codegen();

                // Determine is value is global or local (on stack)
                if (llvm::isa<llvm::GlobalVariable>(variableValue)) {
                    llvm::Constant* constVal =
                        llvm::cast<llvm::Constant>(initializerCode);
                    llvm::GlobalVariable* globalDecl =
                        llvm::cast<llvm::GlobalVariable>(variableValue);
                        globalDecl->setInitializer(constVal);
                }
                else {
                    bool isVolatile =
                            (declType.quality == TypeQualityVal::VOLATILE ||
                            declType.quality == TypeQualityVal::VOLATILE_CONST);

                    // Loading global variable (likely contstant) in stack value
                    if(llvm::isa<llvm::GlobalVariable>(initializerCode)) {
                        // Global variables are pointers to a type
                        llvm::Type* pointeeType = initializerCode->getType()->getContainedType(0);

                        // Create memcpy using data layout information to
                        // fill in size/alignment info
                        auto dataLayout = llvm::DataLayout(LLVMModule.get());
                        Builder.CreateMemCpy(variableValue,
                            variableValue->getPointerAlignment(dataLayout),
                            initializerCode,
                            initializerCode->getPointerAlignment(dataLayout),
                            dataLayout.getTypeStoreSize(pointeeType),
                            isVolatile);

                    }
                    else {
                        Builder.CreateStore(initializerCode,
                                            variableValue,
                                            isVolatile);
                    }

                }
            }

            // Add value to symbol table to be referenced later
            codegenTable.insert(SymbolValue(declName,
                                            variableValue));

            // Clean up constructed decl Value
            varInstruction->deleteValue();

            // return variableValue;
        }
    }
    return nullptr;
}

QuantumValue DeclarationNode::quantum_codegen(QuantumContext& context) {
    int declIdIndex = 0;
    for (auto& initDecl : initList) {
        if(id_info[declIdIndex].type.get_base_type_storage() ==
            StorageSpecVal::QUANTUM) {
            context.assignQubitAddress(initDecl->id_name, id_info[declIdIndex].type);
        }
        else {
            context.declareClassicalMemory(initDecl->id_name, id_info[declIdIndex].type);
        }
        initDecl->quantum_codegen(context);
        declIdIndex++;
    }
    return QUANTUM_NULL_VAL;
}

bool DeclarationNode::classof(const ASTNode * node) {
    auto nodeKind = node->getNodeKind();
    return nodeKind == ASTNodeKind::AST_DeclarationNode;
}
// class DeclarationNode

InitDeclaratorNode::InitDeclaratorNode(std::shared_ptr<DeclaratorNode> decl,
                                       int line, int col) :
DeclNode(ASTNodeKind::AST_InitDeclaratorNode, line, col) {
    this->decl = decl;
    this->hasInitalizer = false;
    this->initializer = ExprNode::NULL_VAL;
    this->id_name = decl->id_name;

}

InitDeclaratorNode::InitDeclaratorNode(std::shared_ptr<DeclaratorNode> decl,
                                       std::shared_ptr<ExprNode> initializer,
                                       int line, int col) :
DeclNode(ASTNodeKind::AST_InitDeclaratorNode, line, col) {
    this->decl = decl;
    this->hasInitalizer = true;
    this->initializer = initializer;
    this->id_name = decl->id_name;
}

void InitDeclaratorNode::unparse(int indent) {
    decl->unparse();
    std::cout << " = ";
    initializer->unparse();
}

void InitDeclaratorNode::checkSemantics() {
    // Record the curent scope level and check the declarator
    int initial_scope_depth = semanticTable.scopeDepth();
    decl->checkSemantics();
    try {
        SymbolInfo& id = (semanticTable.scopeDepth() > initial_scope_depth) ?
                          semanticTable.globalLookup(id_name) :
                          semanticTable.localLookup(id_name);

        // check semantics of initializer expression
        if (hasInitalizer) {
            initializer->checkSemantics();
            initializerType = initializer->type;
        }

        auto declPtr = decl->direct_decl.get();
        if (llvm::isa<IdentifierDeclaratorNode>(declPtr)) {
            if (hasInitalizer) {
                id.kind.value = KindValue::INITVAR;
            }
            else {
                id.kind.value = KindValue::UNINITVAR;
            }
        }
        else if(llvm::isa<FunctionDeclaratorNode>(declPtr)) {
            // Check that an initializer wasn't used
            if (hasInitalizer) {
                mustBe(false, error() + "function '" + id_name +
                              "' is initialized like a variable");
            }
        }

        // Close out extra scopes some declarations (primarily functions)
        // open when checking
        if (semanticTable.scopeDepth() > initial_scope_depth) {
            for (int i = 0;
                   i < semanticTable.scopeDepth() - initial_scope_depth;
                   i++) {
                try {
                    semanticTable.closeScope();
                }
                catch (EmptySTException& emptyEx) {
                    // Shouldn't occur
                    mustBe(false, error() + "Unexpected failure while checking program; this must be a compiler bug.");
                    break;
                 }
            }
        }
    }
    catch (FailedLookupException &lookupEx) {
        // Shouldn't occur
        mustBe(false, error() + "Unexpected failure while checking program; this must be a compiler bug.");
    }
}

llvm::Value* InitDeclaratorNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    auto declCode = decl->codegen(parent, typeInfo);

    return declCode;
}

QuantumValue InitDeclaratorNode::quantum_codegen(QuantumContext& context) {
    // TODO: handle making sure declares of classical memory are at top of program
    // decl->quantum_codegen(context);
    // auto qubits = context.getQubitAddressRange(id_name);
    // TODO: handle more complicated expressions, classic types, and initialization
    /*
    context.programString += "MOVE " + decl->direct_decl->id_name + " ";
    initializer->quantum_codegen(context);
    context.programString += ";\n";
    */
    auto declValue = context.getQuantumValue(id_name);
    auto initVal = initializer->quantum_codegen(context);

    if (declValue.isClassical && !initVal.isNull) {
        context.buildMove(declValue, initVal);
    }
    else if (!initVal.isNull) {
        context.buildQuantumInit(declValue, initVal);
    }
    return QUANTUM_NULL_VAL;
}

bool InitDeclaratorNode::classof(const ASTNode * node) {
    auto nodeKind = node->getNodeKind();
    return nodeKind == ASTNodeKind::AST_InitDeclaratorNode;
}
// class DeclarationNode

/*
VariableDeclNode::VariableDeclNode(ASTNodeKind nodeKind) : DeclNode::DeclNode(nodeKind) {}
VariableDeclNode::VariableDeclNode(ASTNodeKind nodeKind, int line, int col) : DeclNode::DeclNode(nodeKind, line, col) {}
bool VariableDeclNode::classof(const ASTNode * node) {
    auto nodeKind = node->getNodeKind();
    return nodeKind >= ASTNodeKind::AST_VariableDeclNode &&
           nodeKind <= ASTNodeKind::AST_ArrayDeclNode;
} // classof
// class VariableDeclNode


VarDeclNode::VarDeclNode(IdentNode varName, std::shared_ptr<TypeNode> varType, std::shared_ptr<ExprNode> initValue, int line, int col) : VariableDeclNode::VariableDeclNode(ASTNodeKind::AST_VarDeclNode, line, col) {
    this->varName = varName;
    this->varType = varType;
    this->initValue = initValue;
}

void VarDeclNode::unparse(int indent) {
    std::cout << std::left << std::setw(4) << linenum << ":";
    genIndent(indent);
    std::cout << "var ";
    varName.unparse();
    if (varType->isNull()) {
        std::cout << " = ";
        initValue->unparse();
    }
    else {
        std::cout << " ";
        varType->unparse();
    }
    std::cout << " ;" << "\n";
}

void VarDeclNode::checkSemantics() {
    // Check init value is valid type, if present
    initValue->checkSemantics();
    if (!initValue->isNull()) {
        std::string message = "Invalid expression type for variable declaration.";
        std::vector<TypeValue> badTypeList = {TypeValue::UNKNOWN, TypeValue::VOID};
        typeMustNotBe(initValue->type.value, badTypeList,
                      error() + message);
    }

    // Check if variable name has been previously defined in same scope
    SymbolInfo id;
    try {
        id = semanticTable.localLookup(varName.idname);

        // The prior should throw a FailedLookupException before this
        // If not, then the variable name has already been declared
        mustBe(false, error() + id.get_name() + " is already declared.");
        varName.type = Type(TypeValue::ERROR);
    }
    catch (...) {

        // Pick whether the variable is initialized or not based on presence
        // of an expression and if it is a global variable
        bool isGlobal = semanticTable.scopeDepth() == 1;
        Kind varKind = (initValue->isNull() && !isGlobal) ?
                        Kind(KindValue::UNINITVAR) : Kind(KindValue::INITVAR);
        // Pick type info for identifier name based on presence of explicit type
        id = SymbolInfo(varName.idname, varKind, (varType->isNull()) ?
                                                       initValue->type :
                                                       varType->type);


        try {
            semanticTable.insert(id);
        } catch (DuplicateException &dupEx){
            // can't happen
        } catch (EmptySTException &emptyEx){
            // can't happen
        }

        // Ensure name gets correct type information
        varName.checkSemantics();
    }
} // checkSemantics

// class VarDeclNode

ConstDeclNode::ConstDeclNode(IdentNode constName, std::shared_ptr<ExprNode> constValue, int line, int col)
: VariableDeclNode::VariableDeclNode(ASTNodeKind::AST_ConstDeclNode, line, col) {
    this->constName = constName;
    this->constValue = constValue;
}

void ConstDeclNode::unparse(int indent) {
    std::cout << std::left << std::setw(4) << linenum << ":";
    genIndent(indent);
    std::cout << "const ";
    constName.unparse();
    std::cout << " = ";
    constValue->unparse();
    std::cout << " ;" << "\n";
}

bool ConstDeclNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_ConstDeclNode;
} // classof

void ConstDeclNode::checkSemantics() {

    // Check const value is valid
    constValue->checkSemantics();
    std::string message = "Invalid expression type for constant declaration.";
    std::vector<TypeValue> badTypeList = {TypeValue::UNKNOWN, TypeValue::VOID};
    typeMustNotBe(constValue->type.value, badTypeList,
                  error() + message);

    // Check if const name has been previously defined in same scope
    SymbolInfo id;
    try {
        id = semanticTable.localLookup(constName.idname);

        // Prior statement should throw a FailedLookupException
        // If
        mustBe(false, error() + id.get_name() + " is already declared.");
        constName.type = Type(TypeValue::ERROR);
    }
    catch (...) {

        // Infer type info for const name from given value expression
        id = SymbolInfo(constName.idname, Kind(KindValue::CONST), constValue->type);

        try {
            semanticTable.insert(id);
        } catch (DuplicateException &dupEx){
            // can't happen
        } catch (EmptySTException &emptyEx){
            // can't happen
        }

        // Ensure name gets correct type information
        constName.checkSemantics();
    }

} // checkSemantics

// class ConstDeclNode

ArrayDeclNode::ArrayDeclNode(IdentNode arrayName, std::shared_ptr<TypeNode> elementType, IntLitNode arraySize, int line, int col) : VariableDeclNode::VariableDeclNode(ASTNodeKind::AST_ArrayDeclNode, line, col) {
    this->arrayName = arrayName;
    this->elementType = elementType;
    this->arraySize = arraySize;
}

void ArrayDeclNode::unparse(int indent) {
    std::cout << std::left << std::setw(4) << linenum << ":";
    genIndent(indent);
    std::cout << "var ";
    arrayName.unparse();
    std::cout << " ";
    elementType->unparse();
    std::cout << " [ ";
    arraySize.unparse();
    std::cout << " ] ;" << "\n";
}

bool ArrayDeclNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_ArrayDeclNode;
} // classof

void ArrayDeclNode::checkSemantics() {

    // Check if variable name has been previously defined in same scope
    SymbolInfo id;
    try {
        id = semanticTable.localLookup(arrayName.idname);

        // Lookup should throw an error before this.
        // If not, the variable has been previously defined; mark as an error
        mustBe(false, error() + id.get_name() + " is already declared.");
        arrayName.type = Type(TypeValue::ERROR);
    }
    catch (...) {

        // Check for valid array size; use a placeholder size of 1 if not valid
        mustBe(arraySize.intval > 0, error() + "Sizes of arrays must be greater than 0.");
        int size = (arraySize.intval > 0) ? arraySize.intval : 1;

        // Pick type info for identifier name based on explicit type and size
        id = SymbolInfo(arrayName.idname, Kind(KindValue::INITVAR),
                            Type(elementType->type.value, size));

        try {
            semanticTable.insert(id);
        } catch (DuplicateException &dupEx){
            // can't happen
        } catch (EmptySTException &emptyEx){
            // can't happen
        }

        // Ensure name gets correct type information
        arrayName.checkSemantics();
    }

} // checkSemantics

// class ArrayDeclNode
*/

FuncDefNode::FuncDefNode(DeclSpecifierList returnType, std::shared_ptr<DeclaratorNode> declarator, std::shared_ptr<BlockNode> body, int line, int col) : DeclNode::DeclNode(ASTNodeKind::AST_FuncDefNode, line, col) {
    this->returnType = returnType;
    this->declarator = declarator;
    this->body = body;
}

void FuncDefNode::unparse(int indent) {
    std::cout << std::left << std::setw(4) << linenum << ":";
    genIndent(indent);
    for (auto& type_specs : returnType) {
        type_specs->unparse();
        std::cout << " ";
    }
    std::cout << " ";
    declarator->unparse();
    std::cout << "\n";
    body->unparse(indent);
}

type_info::Type FuncDefNode::generate_return_type() {
    Type returnTypeSpecs = Type();
    returnTypeSpecs.isScalar = true;
    bool set_signedness = false;
    bool set_storage = false;
    bool set_quality = false;
    for (auto& spec : returnType) {
        auto type_node = spec.get();
        if (llvm::isa<TypeSpecNode>(type_node)) {
            if (llvm::isa<LongTypeNode>(type_node)) {
                if (returnTypeSpecs.value == TypeValue::UNKNOWN ||
                    returnTypeSpecs.value == TypeValue::INTEGER) {
                    returnTypeSpecs.value = TypeValue::LONG;
                }
                else if (returnTypeSpecs.value == TypeValue::LONG) {
                    returnTypeSpecs = TypeValue::LONGLONG;
                }
                else {
                    mustBe(false, error() + "both 'long' and '" +
                           returnTypeSpecs.get_value_string() +
                           "' in declaration specifiers");
                    return Type(TypeValue::ERROR);
                }
            }
            else if (llvm::isa<ShortTypeNode>(type_node)) {
                if (returnTypeSpecs.value == TypeValue::UNKNOWN ||
                    returnTypeSpecs.value == TypeValue::INTEGER) {
                    returnTypeSpecs.value = TypeValue::SHORT;
                }
                else if (returnTypeSpecs.value == TypeValue::SHORT) {
                    mustBe(false, error() + "duplicate 'short'");
                    return Type(TypeValue::ERROR);
                }
                else {
                    mustBe(false, error() + "both 'short' and '" +
                           returnTypeSpecs.get_value_string() +
                           "' in declaration specifiers");
                    return Type(TypeValue::ERROR);
                }
            }
            else if (llvm::isa<SignedTypeNode>(type_node)) {
                if (set_signedness) {
                    if (returnTypeSpecs.isSigned) {
                        mustBe(false, error() + "duplicate 'signed'");
                    }
                    else {
                        mustBe(false, error() + "both 'signed' and 'unsigned' in declaration specifiers");
                    }
                    return Type(TypeValue::ERROR);
                } else {
                    returnTypeSpecs.isSigned = true;
                    set_signedness = true;
                }
            }
            else if (llvm::isa<UnsignedTypeNode>(type_node)) {
                if (set_signedness) {
                    if (!returnTypeSpecs.isSigned) {
                        mustBe(false, error() + "duplicate 'unsigned'");
                    }
                    else {
                        mustBe(false, error() + "both 'signed' and 'unsigned' in declaration specifiers");
                    }
                    return Type(TypeValue::ERROR);
                } else {
                    returnTypeSpecs.isSigned = false;
                    set_signedness = true;
                }
            }
            else if (returnTypeSpecs.value == TypeValue::UNKNOWN) {
                returnTypeSpecs.value = spec->type.value;
            }
            else {
                mustBe(false, error() +
                       "two or more data types in declaration specifiers");
                return Type(TypeValue::ERROR);
            }
        }
        else if (llvm::isa<StorageClassSpecNode>(type_node)) {
            if (set_storage) {
                mustBe(false, error() +
                       "multiple storage classes in declaration specifiers");
                return Type(TypeValue::ERROR);
            }
            else if (llvm::isa<ExternNode>(type_node)) {
                returnTypeSpecs.storage = StorageSpecVal::EXTERN;
                set_storage = true;
            }
            else if (llvm::isa<StaticNode>(type_node)) {
                returnTypeSpecs.storage = StorageSpecVal::STATIC;
                set_storage = true;
            }
            else if (llvm::isa<QuantumNode>(type_node)) {
                returnTypeSpecs.storage = StorageSpecVal::QUANTUM;
                set_storage = true;
            }
            else {
                mustBe(false, error() +
                    "function definition not declared as 'extern', 'static', or 'quantum'");
                return Type(TypeValue::ERROR);
            }
        }
        else if (llvm::isa<TypeQualifierNode>(type_node)) {
            set_quality = true;
            if (returnTypeSpecs.quality == TypeQualityVal::VOLATILE_CONST) {
                continue; // Any additional qualifiers are ignored
            }
            else if (llvm::isa<VolatileNode>(type_node)) {
                returnTypeSpecs.quality =
                    (returnTypeSpecs.quality == TypeQualityVal::CONST) ?
                    TypeQualityVal::VOLATILE_CONST :
                    TypeQualityVal::VOLATILE;

            } else { // llvm::isa<ConstNode>(spec)
                returnTypeSpecs.quality =
                    (returnTypeSpecs.quality == TypeQualityVal::VOLATILE) ?
                    TypeQualityVal::VOLATILE_CONST :
                    TypeQualityVal::CONST;
            }
        }
        else {
            mustBe(false, error() + "Unimplemented type used");
            return Type(TypeValue::ERROR);
        }
    }

    // Establish defaults if left implicit in the type specs
    if (returnTypeSpecs.value == TypeValue::UNKNOWN) {
        returnTypeSpecs.value = TypeValue::INTEGER;
    }
    if (!set_signedness) {
        returnTypeSpecs.isSigned = true;
    }
    if (!set_storage) {
        returnTypeSpecs.storage = StorageSpecVal::EXTERN;
    }
    if (!set_quality) {
        returnTypeSpecs.quality = TypeQualityVal::NONE;
    }
    return returnTypeSpecs;
}

bool FuncDefNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_FuncDefNode;
} // classof

void FuncDefNode::checkSemantics() {
    // Keep track of current scope depth
    int initial_scope_depth = semanticTable.scopeDepth();

     // Check declarator is valid and fetch information
     declarator->checkSemantics();
     try {
         SymbolInfo& func_id = semanticTable.globalLookup(declarator->id_name);

         // Confirm that the declartor was in fact a function declarator
         mustBe(func_id.type.isFunction, error() + "Unexpected declarator for the function '" + declarator->id_name + "'");

         // Generate and update return type in symbol table
         auto returnType = std::make_shared<Type>();
         *returnType = generate_return_type();
         if (returnType->get_base_type_storage() == StorageSpecVal::QUANTUM) {
            numQuantumFuncs++;
         }
         func_id.type = Type(returnType, func_id.type.functionArgTypes,
                             func_id.type.isVarArg,
                             returnType->quality, returnType->storage);
         id_info = func_id;

         // Check semantics of body
         // Scope should already be open with arugments included

         body->setGivenScope(true);
         body->setIsFuncBody();
         bool expectReturn = !(returnType->value == TypeValue::VOID &&
                               returnType->isScalar);
         body->setReturnStatus(expectReturn,
                               *returnType);
         body->checkSemantics();
     } catch (FailedLookupException &lookupEx) {
         // Shouldn't occur
         mustBe(false, "Unexpected failure while checking program; this must be a compiler bug.");
     }

     // close function scope for arguments
     // Close out extra scopes some declarations (primarily arguments)
     // opened when checking
     if (semanticTable.scopeDepth() > initial_scope_depth) {
         for (int i = 0;
                i < semanticTable.scopeDepth() - initial_scope_depth;
                i++) {
             try {
                 semanticTable.closeScope();
             }
             catch (EmptySTException& emptyEx) {
                 // Shouldn't occur
                 mustBe(false, error() + "Unexpected failure while checking program; this must be a compiler bug.");
                 break;
              }
         }
     }

    /*
    // Check if variable name has been previously defined in same scope
    SymbolInfo id;
    try {
        id = semanticTable.localLookup(name.idname);

        // Lookup should throw an error before this.
        // If not, the variable has been previously defined; mark as an error
        mustBe(false, error() + id.get_name() + " is already declared.");
        name.type = Type(TypeValue::ERROR);
    }
    catch (...) {

        // Pick type info for identifier name based on return type
        Type "Unexpected failure while checking program; this must be a compiler bug." = Type(returnType->type.value, std::vector<Type>());
        id = SymbolInfo(name.idname, Kind(KindValue::FUNC), funcType);

        try {
            semanticTable.insert(id);
        } catch (DuplicateException &dupEx){
            // can't happen
        } catch (EmptySTException &emptyEx){
            // can't happen
        }

        // Ensure name gets correct type information
        name.checkSemantics();

        // Check arguments and add argument types to function type signature
        semanticTable.openScope();
        for (auto arg : args) {
            arg->checkSemantics();
            funcType.functionArgTypes.push_back(arg->argName.type);
        }

        // Setup block to check for return statements as needed
        body.setGivenScope(true);
        bool returnCondition = (returnType->type.value == TypeValue::VOID) ?
                              BlockNode::returnNotNeccesary :
                              BlockNode::returnRequired;
        body.setReturnStatus(returnCondition, returnType->type.value);
        body.setIsFuncBody();
        body.checkSemantics();
        try {
            semanticTable.closeScope();
        } catch (EmptySTException &emptyEx){
            // can't happen
        }

    }
    */
} // checkSemantics

llvm::Value* FuncDefNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    if(id_info.type.storage == StorageSpecVal::QUANTUM) {
        // Gen for classical args: "DECLARE temp_<func_name>_arg_<x> <TYPE>"
        // "DEFGATE <FUNC_NAME> [<quantum_arg_names>] :"
        // std::quil_instructions = body.quantum_codegen(); TODO: define signature of method
        // Insert tab before every instruction line
        // Return values just left off for quantum return values (probably)?
        // "Declare <func_name>_return <TYPE>" if returning classical
        // Create string constant for LLVM (See string constant node)
        // String naming convention: __quant_<func_name>

        // At this point, just get the program body and use that as the program
        // Store it as a string literal in the code, using the name of the function for later lookup when called
        QuantumContext func_context = QuantumContext(qubitOffset);
        func_context.setFunctionName(id_info.get_name());
        auto funcDeclarator = llvm::cast<FunctionDeclaratorNode>(declarator->direct_decl.get());
        auto argNames = funcDeclarator->argNames;

        int argNameIndex = 0;
        for (auto& type: id_info.type.functionArgTypes) {
            auto name = argNames[argNameIndex];
            func_context.addFunctionArgument(name, type);
            argNameIndex++;
        }
        func_context.setFunctionReturnType(*(id_info.type.subType));

        body->quantum_codegen(func_context);

        quilCode += func_context.genQUIL();
        qubitOffset = func_context.nextUnassignedQubit;

        numQuantumFuncs--;
        if (numQuantumFuncs == 0) {
            // Generate final strirng once all quantum functions generated
            // Creat global variable for string, then assign to global variable
            std::string quantum_var_name = std::string(".str.quantc");

            llvm::Constant* func_string_val = llvm::ConstantDataArray::getString(
                                                LLVMContext,
                                                quilCode,
                                                true
            );

            llvm::Type* func_string_type =
                llvm::Type::getIntNPtrTy(LLVMContext, sizeof(char)*CHAR_BIT);

            // If variables previously defined by prior function, erase them
            // to create updated versions
            /*
            if (!firstQuantFunc) {
                llvm::GlobalVariable * quant_string_var =
                    LLVMModule->getNamedGlobal(quantum_var_name + ".tmp");
                llvm::GlobalVariable * quant_var =
                    LLVMModule->getNamedGlobal(quantum_var_name);
                quant_string_var->setInitializer(nullptr);
                quant_string_var->eraseFromParent();
                quant_var->setInitializer(nullptr);
                quant_var->eraseFromParent();
            }
            */

            LLVMModule->getOrInsertGlobal(
                            quantum_var_name + ".tmp",
                            func_string_val->getType());
            llvm::GlobalVariable * quant_string_var =
                LLVMModule->getNamedGlobal(quantum_var_name + ".tmp");
            quant_string_var->setLinkage(llvm::GlobalValue::PrivateLinkage);
            quant_string_var->setAlignment(llvm::Align(sizeof(char)*CHAR_BIT));
            quant_string_var->setInitializer(func_string_val);
            quant_string_var->setConstant(true);

            LLVMModule->getOrInsertGlobal(
                            quantum_var_name,
                            func_string_type);
            llvm::GlobalVariable * quant_var =
                LLVMModule->getNamedGlobal(quantum_var_name);
            quant_var->setLinkage(llvm::GlobalValue::InternalLinkage);
            quant_var->setAlignment(llvm::Align(sizeof(char)*CHAR_BIT));
            auto cast = llvm::ConstantExpr::getBitCast(quant_string_var, func_string_type);
            quant_var->setInitializer(cast);
            quant_var->setConstant(true);
            firstQuantFunc = false;
        }

        return nullptr;

    } else {
        // Get incomplete function type from declarator
        // (includes name and argument type/names)
        llvm::Function* declCode = llvm::cast<llvm::Function>(declarator->codegen());

        // Generate LLVM Function by combining
        // 1) Previous return type info
        // 2) Basic name + arg information from the declarator
        llvm::FunctionType* func_type =
            llvm::cast<llvm::FunctionType>(id_info.type.codegen_type(LLVMContext));

        auto linkage = (id_info.type.storage == StorageSpecVal::EXTERN) ?
                        llvm::Function::ExternalLinkage :
                        llvm::Function::PrivateLinkage;
        llvm::Function* functionCode = llvm::Function::Create(func_type,
                                                               linkage,
                                                               declarator->id_name,
                                                               LLVMModule.get());
        functionCode->setCallingConv(llvm::CallingConv::C);

        // Open new scope for argument values
        codegenTable.openScope();

        // Set argument names in function code from declarator
        // and capture the argument values in the symbol values table
        llvm::Function::arg_iterator declArgIter = declCode->arg_begin();
        for (llvm::Function::arg_iterator funcdefArgIter = functionCode->arg_begin();
             funcdefArgIter != functionCode->arg_end();
             funcdefArgIter++) {
                auto argName = declArgIter->getName();
                funcdefArgIter->setName(argName);
                codegenTable.insert(SymbolValue(argName, funcdefArgIter));
                declArgIter++;
        }

        // Generate code for the body
        /*
        NamedValues.clear();
        for (auto& arg: functionCode->args()) {
            NamedValues[arg.getName()] = &arg;
        }
        */

        // Open new scope for body
        codegenTable.openScope();
        if (body->codegen(functionCode)) {
            llvm::BasicBlock* endOfFunc = Builder.GetInsertBlock();
            if (!endOfFunc->getTerminator()) {
                Builder.CreateUnreachable();
            }
            llvm::verifyFunction(*functionCode);
            codegenTable.closeScope();
            codegenTable.closeScope();
            return functionCode;
        }
        else {
            // Error generating body code
            functionCode->eraseFromParent();
            codegenTable.closeScope();
            codegenTable.closeScope();
            return nullptr;
        }
    }

}
// class FuncDefNode



DeclaratorNode::DeclaratorNode(
    PointerList pointer_spec,
    std::shared_ptr<DirectDeclaratorNode> direct_decl,
    int line, int col) :
DeclNode::DeclNode(ASTNodeKind::AST_DeclaratorNode, line, col) {
    this->pointer_spec = pointer_spec;
    this->direct_decl = direct_decl;
    this->id_name = direct_decl->id_name;
}

void DeclaratorNode::unparse(int indent) {
    for (auto& pointer : pointer_spec) {
        pointer->unparse();
        std::cout << " ";
    }
    direct_decl->unparse();
}

bool DeclaratorNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_DeclaratorNode;
} // classof

void DeclaratorNode::checkSemantics() {
    // Keep track of initial scope in case sub declarator(s) are function decls
    int initial_scope_depth = semanticTable.scopeDepth();
    direct_decl->checkSemantics();

    // Fetch the current type information from the symbol table
    try {

        SymbolInfo& id = (semanticTable.scopeDepth() > initial_scope_depth) ?
                          semanticTable.globalLookup(direct_decl->id_name) :
                          semanticTable.localLookup(direct_decl->id_name);

        // Modify symbol's type information to reflect pointer information
        for (auto& pointer : pointer_spec) {

            // Determine pointer qualifiers
            bool isVolatile = (std::find_if(pointer->qualifiers.begin(),
                                    pointer->qualifiers.end(),
                                    [&](std::shared_ptr<TypeQualifierNode> q){
                                        return llvm::isa<VolatileNode>(q.get());
                                    })
                                    != pointer->qualifiers.end());
            bool isConst = (std::find_if(pointer->qualifiers.begin(),
                                    pointer->qualifiers.end(),
                                    [&](std::shared_ptr<TypeQualifierNode> q){
                                        return llvm::isa<ConstNode>(q.get());
                                    })
                                    != pointer->qualifiers.end());
            bool isBoth = isVolatile && isConst;
            TypeQualityVal pointer_qual;
            if (isBoth) {
                pointer_qual = TypeQualityVal::VOLATILE_CONST;
            }
            else if (isVolatile) {
                pointer_qual = TypeQualityVal::VOLATILE;
            }
            else if (isConst) {
                pointer_qual = TypeQualityVal::CONST;
            }
            else {
                pointer_qual = TypeQualityVal::NONE;
            }

            // Determine how to modify symbol's type information
            if (id.type.isFunction) {
                // Set return type to be a pointer to current return type
                auto return_type = std::make_shared<Type>(id.type.subType,
                                                           pointer_qual);
                id.type = Type(return_type, id.type.functionArgTypes,
                               id.type.isVarArg,
                               id.type.quality,
                               id.type.storage);
            }
            else if (id.type.isArray) {
                // Set element type to be a pointer to current element type
                auto element_type = std::make_shared<Type>(id.type.subType,
                                                           pointer_qual);
                id.type = Type(element_type, id.type.size,
                               id.type.quality,
                               id.type.storage);
            }
            else if (id.type.isPointer) {
                // Set type to be pointer to current pointer type
                auto pointee_type = std::make_shared<Type>(id.type);
                id.type = Type(pointee_type, pointer_qual);
            }
            else {
                // Set type to be void pointer
                // (specific type info isn't available at this point)
                auto void_type = std::make_shared<Type>(TypeValue::VOID);
                id.type = Type(void_type, pointer_qual);
            }
        }

    }
    catch (FailedLookupException &lookupEx) {
        // Shouldn't occur
        mustBe(false, error() + "Unexpected failure while checking program; this must be a compiler bug.");
    }
} // checkSemantics

llvm::Value* DeclaratorNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    // TODO: generate code for case when this is a pointer
    return direct_decl->codegen(parent, typeInfo);
}

// TODO: Remove?
QuantumValue DeclaratorNode::quantum_codegen(QuantumContext& context) {
    // Ignore pointer information; should be invalid in this case
    // TODO: make sure this is enough
    return QUANTUM_NULL_VAL;
}
// class DeclaratorNode


DirectDeclaratorNode::DirectDeclaratorNode(ASTNodeKind nodeKind, int line, int col) : DeclNode::DeclNode(nodeKind, line, col) {}
bool DirectDeclaratorNode::classof(const ASTNode * node) {
    auto nodeKind = node->getNodeKind();
    return nodeKind >= ASTNodeKind::AST_DirectDeclaratorNode &&
           nodeKind <= ASTNodeKind::AST_FunctionDeclaratorNode;
} // classof
// class DirectDeclaratorNode


IdentifierDeclaratorNode::IdentifierDeclaratorNode(std::string name,
    int line, int col) :
DirectDeclaratorNode::DirectDeclaratorNode(ASTNodeKind::AST_IdentifierDeclaratorNode, line, col) {
    this->id_name = name;
}

void IdentifierDeclaratorNode::unparse(int indent) {
    std::cout << id_name;
}

bool IdentifierDeclaratorNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_IdentifierDeclaratorNode;
} // classof

void IdentifierDeclaratorNode::checkSemantics() {
    // Attempt to add identifier to the symbol table
    try {
        semanticTable.insert(SymbolInfo(id_name, Kind(), Type()));
    }
    catch (EmptySTException& emptyEx) {
        mustBe(false, "Unexpected failure while checking program; this must be a compiler bug.");
    }
    catch (DuplicateException& dupEx) {
        mustBe(false, error() + "redefinition of '" + id_name + "'.");
    }


} // checkSemantics

llvm::Value* IdentifierDeclaratorNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    // Will fill in type information later
    return new llvm::AllocaInst(typeInfo, 0, nullptr, llvm::Align(), id_name);
} // codegen

// No code_gen needed
// class IdentifierDeclaratorNode



ParensDeclaratorNode::ParensDeclaratorNode(
    std::shared_ptr<DeclaratorNode> decl,
    int line, int col)
: DirectDeclaratorNode::DirectDeclaratorNode(ASTNodeKind::AST_ParensDeclaratorNode, line, col) {
    this->decl = decl;
    this->id_name = decl->id_name;
}

void ParensDeclaratorNode::unparse(int indent) {
    std::cout << " ( ";
    decl->unparse();
    std::cout << " ) ";
}

bool ParensDeclaratorNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_ParensDeclaratorNode;
} // classof

void ParensDeclaratorNode::checkSemantics() {
 // Nothing to do other than pass check onto further decl
 decl->checkSemantics();
} // checkSemantics

llvm::Value* ParensDeclaratorNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    return decl->codegen(parent, typeInfo);
}

QuantumValue ParensDeclaratorNode::quantum_codegen(QuantumContext& context) {
    return decl->quantum_codegen(context);
}

// class ParensDeclaratorNode


ArrayDeclaratorNode::ArrayDeclaratorNode(
    std::shared_ptr<DirectDeclaratorNode> decl,
    std::shared_ptr<ExprNode> size,
    int line, int col)
: DirectDeclaratorNode::DirectDeclaratorNode(ASTNodeKind::AST_ArrayDeclaratorNode, line, col) {
    this->decl = decl;
    this->size = size;
    this->id_name = decl->id_name;
}

void ArrayDeclaratorNode::unparse(int indent) {
    decl->unparse();
    std::cout << " [ ";
    size->unparse();
    std::cout << " ] ";
}

bool ArrayDeclaratorNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_ArrayDeclaratorNode;
} // classof

void ArrayDeclaratorNode::checkSemantics() {
    // NOTE: semantic check for function returning array occurs here ?
    decl->checkSemantics();
    size->checkSemantics();

    try {
        SymbolInfo& arrayName = semanticTable.localLookup(decl->id_name);

        // Update type to be an array value
        // use 0 to indicate unknown size
        // auto typePtr = std::make_shared<Type>();
        // *typePtr = arrayName.type;

        int arraySize = 0;
        auto sizePtr = size.get();
        if (llvm::isa<IntLitNode>(sizePtr)) {
            auto intLitNode = llvm::cast<IntLitNode>(sizePtr);
            arraySize = intLitNode->intval;
        }
        // arrayName.type = Type(typePtr, arraySize);
        auto type = Type(std::make_shared<Type>(TypeValue::VOID), arraySize);
        arrayName.type.set_base_type(type);
        if (arrayName.kind.value == KindValue::OTHER) {
            arrayName.kind = Kind(KindValue::UNINITVAR);
        }

    }
    catch (FailedLookupException& lookupEx) {
        mustBe(false, "Unexpected failure while checking program; this must be a compiler bug.");
    }
} // checkSemantics

llvm::Value* ArrayDeclaratorNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    if (typeInfo->isArrayTy()) {
        auto arrayType = llvm::cast<llvm::ArrayType>(typeInfo);

        llvm::Value* declVal = decl->codegen(nullptr,
                                             arrayType->getElementType());
        llvm::AllocaInst* declAlloc = llvm::cast<llvm::AllocaInst>(declVal);

        // If size is know at compile time, allocate as an array type directly
        if (arrayType->getNumElements() > 0) {
            auto fixedArrayType = llvm::ArrayType::get(
                                        declAlloc->getAllocatedType(),
                                        arrayType->getNumElements());
            return new llvm::AllocaInst(fixedArrayType, 0, Builder.getInt32(1));
        }
        else {
            llvm::Value* sizeVal = size->codegen();
            sizeVal = Builder.CreateMul(declAlloc->getArraySize(), sizeVal);

            return new llvm::AllocaInst(declAlloc->getAllocatedType(),
                                        0, // default address space
                                        sizeVal,
                                        llvm::Align(declAlloc->getAlignment()),
                                        id_name);
        }
    }
    else {
        // Error
        return nullptr;
    }

}
// class ArrayDeclaratorNode


FunctionDeclaratorNode::FunctionDeclaratorNode(
    std::shared_ptr<DirectDeclaratorNode> decl,
    ArgumentList arguments,
    bool isVarArg,
    int line, int col) :
DirectDeclaratorNode::DirectDeclaratorNode(
                        ASTNodeKind::AST_FunctionDeclaratorNode, line, col) {
    this->decl = decl;
    this->arguments = arguments;
    this->isVarArg = isVarArg;
    this->id_name = decl->id_name;
}

void FunctionDeclaratorNode::unparse(int indent) {
    decl->unparse();
    std::cout << " ( ";
    for (auto& arg : arguments) {
        arg->unparse();
        std::cout << " , ";
    }

    // Back up to delete the last comma if arguments are printed
    if (isVarArg) {
        std::cout << "...";
    }
    else if (arguments.size() > 0) {
        std::cout << "\b\b\b";
    }
    std::cout << " ) ";
}

bool FunctionDeclaratorNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_FunctionDeclaratorNode;
} // classof

void FunctionDeclaratorNode::checkSemantics() {
    // Attempt to add the symbol to symbol table
    decl->checkSemantics();
    try {
        SymbolInfo& func_name = semanticTable.localLookup(decl->id_name);

        // Check that the arguments are valid and
        // gather their type information
        semanticTable.openScope();
        std::vector<Type> arg_types;
        for (auto& arg: arguments) {
            arg->checkSemantics();
            auto agrPtr = arg.get();

            // Verify that name argument's were inserted into the semanticTable
            if (llvm::isa<NamedArgDeclNode>(agrPtr)) {
                try {
                    NamedArgDeclNode* namedPtr =
                        llvm::cast<NamedArgDeclNode>(agrPtr);
                    auto argName = namedPtr->decl->id_name;
                    argNames.push_back(argName);
                    semanticTable.localLookup(argName);
                }
                catch (FailedLookupException& lookupEx) {
                    mustBe(false, "Unexpected failure while checking program; this must be a compiler bug.");
                }
            }
            else {
                argNames.push_back("");
            }
            arg_types.push_back(arg->argType);
        }

        // Update kind and type of symbol to be a function
        func_name.kind = Kind(KindValue::FUNC);
        auto return_type = std::make_shared<Type>();
        *return_type = func_name.type;
        func_name.type = Type(return_type, arg_types,
                              isVarArg,
                              func_name.type.quality,
                              func_name.type.storage);
    }
    catch (FailedLookupException& lookupEx) {
        mustBe(false, "Unexpected failure while checking program; this must be a compiler bug.");
    }

} // checkSemantics


llvm::Value* FunctionDeclaratorNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    // Generate argument info types (including argument names)
    std::vector<llvm::Type *> argTypeValues;
    std::vector<std::string> argNames;
    for (auto& parameter : arguments) {
        auto paramPtr = parameter.get();
        if(llvm::isa<NamedArgDeclNode>(paramPtr)) {
            auto namedParameterPtr = llvm::cast<NamedArgDeclNode>(paramPtr);
            argNames.push_back(namedParameterPtr->id_info.get_name());
        }
        else {
            argNames.push_back(std::string(""));
        }
        auto argType = parameter->argType.codegen_type(LLVMContext);
        if (argType->isArrayTy()) {
            // Convert array to pointer for arguments
            argType = llvm::PointerType::getUnqual(argType->getContainedType(0));
        }
        argTypeValues.push_back(argType);
    }

    llvm::FunctionType* basic_func_type =
    llvm::FunctionType::get(llvm::Type::getVoidTy(LLVMContext),
                            argTypeValues,
                            false);

    llvm::Function* basic_func_decl = llvm::Function::Create(basic_func_type,
                                                llvm::Function::ExternalLinkage,
                                                decl->id_name);

    // Insert names into arguments
    llvm::Function::arg_iterator newdeclArgIter = basic_func_decl->arg_begin();
    for (auto& name : argNames) {
        newdeclArgIter->setName(name);
        newdeclArgIter++;
    }
    // Generate declarator info
    return basic_func_decl;

}
// class FunctionDeclaratorNode



DestructureDeclaratorNode::DestructureDeclaratorNode(IdentList ids,
                                                     int line, int col) :
DirectDeclaratorNode::DirectDeclaratorNode(
                        ASTNodeKind::AST_DestructureDeclaratorNode, line, col) {
    this->identifiers = ids;
}

void DestructureDeclaratorNode::unparse(int indent) {
    std::cout << " auto [ ";
    for (auto& id: identifiers) {
        std::cout << id->get_value();
        std::cout << " , ";
    }
    if (identifiers.size() > 0) {
        std::cout << "\b\b\b";
    }
    std::cout << " ] ";
}

bool DestructureDeclaratorNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_DestructureDeclaratorNode;
} // classof



void DestructureDeclaratorNode::checkSemantics() {
    for (auto& id: identifiers) {
        try {
            semanticTable.insert(SymbolInfo(id->get_value(), Kind(), Type()));
        }
        catch (EmptySTException& emptyEx) {
            mustBe(false, "Unexpected failure while checking program; this must be a compiler bug.");
        }
        catch (DuplicateException& dupEx) {
            mustBe(false, error() + "redefinition of '" + id->get_value() + "'.");
        }
    }
} // checkSemantics

// TODO: finish implementing DestructureDeclaratorNode codegen
// class DestructureDeclaratorNode
std::shared_ptr<NullAbstractNode> AbstractDeclaratorNode::NULL_VAL = std::make_shared<NullAbstractNode>();

AbstractDeclaratorNode::AbstractDeclaratorNode(ASTNodeKind nodeKind) : DeclNode(AST_AbstractDeclaratorNode) {} ;

AbstractDeclaratorNode::AbstractDeclaratorNode(PointerList pointer_spec,
               std::shared_ptr<DirectAbstractDeclaratorNode> direct_decl,
               int line, int col) : DeclNode(AST_AbstractDeclaratorNode, line, col) {
            this->pointer_spec = pointer_spec;
            this->direct_decl = direct_decl;
}

void AbstractDeclaratorNode::unparse(int indent) {
    for (auto& pointer: pointer_spec) {
        pointer->unparse();
    }
    std::cout << " ";
    direct_decl->unparse();
}

type_info::Type AbstractDeclaratorNode::generateAbstractType() {
    Type direcDeclType = direct_decl->generateDirectAbstractType();

    for (auto& pointer : pointer_spec) {

        // Determine pointer qualifiers
        bool isVolatile = (std::find_if(pointer->qualifiers.begin(),
                                pointer->qualifiers.end(),
                                [&](std::shared_ptr<TypeQualifierNode> q){
                                    return llvm::isa<VolatileNode>(q.get());
                                })
                                != pointer->qualifiers.end());
        bool isConst = (std::find_if(pointer->qualifiers.begin(),
                                pointer->qualifiers.end(),
                                [&](std::shared_ptr<TypeQualifierNode> q){
                                    return llvm::isa<ConstNode>(q.get());
                                })
                                != pointer->qualifiers.end());
        bool isBoth = isVolatile && isConst;
        TypeQualityVal pointer_qual;
        if (isBoth) {
            pointer_qual = TypeQualityVal::VOLATILE_CONST;
        }
        else if (isVolatile) {
            pointer_qual = TypeQualityVal::VOLATILE;
        }
        else if (isConst) {
            pointer_qual = TypeQualityVal::CONST;
        }
        else {
            pointer_qual = TypeQualityVal::NONE;
        }

        // Determine how to modify symbol's type information
        if (direcDeclType.isFunction) {
            // Set return type to be a pointer to current return type
            auto return_type = std::make_shared<Type>(direcDeclType.subType,
                                                       pointer_qual);
            direcDeclType = Type(return_type, direcDeclType.functionArgTypes,
                           direcDeclType.isVarArg,
                           direcDeclType.quality,
                           direcDeclType.storage);
        }
        else if (direcDeclType.isArray) {
            // Set element type to be a pointer to current element type
            auto element_type = std::make_shared<Type>(direcDeclType.subType,
                                                       pointer_qual);
            direcDeclType = Type(element_type, direcDeclType.size,
                           direcDeclType.quality,
                           direcDeclType.storage);
        }
        else if (direcDeclType.isPointer) {
            // Set type to be pointer to current pointer type
            auto pointee_type = std::make_shared<Type>(direcDeclType);
            direcDeclType = Type(pointee_type, pointer_qual);
        }
        else {
            // Set type to be void pointer
            // (specific type info isn't available at this point)
            auto void_type = std::make_shared<Type>(TypeValue::VOID);
            direcDeclType = Type(void_type, pointer_qual);
        }
    }

    return direcDeclType;
}

void AbstractDeclaratorNode::checkSemantics() {};

bool AbstractDeclaratorNode::classof(const ASTNode * node) {
    return node->getNodeKind() >= ASTNodeKind::AST_AbstractDeclaratorNode &&
           node->getNodeKind() <= ASTNodeKind::AST_NullAbstractNode;
}
// class AbstractDeclaratorNode

NullAbstractNode::NullAbstractNode() : AbstractDeclaratorNode(AST_NullAbstractNode) {}
bool NullAbstractNode::isNull() { return true; }
void NullAbstractNode::unparse(int indent) {};
type_info::Type NullAbstractNode::generateAbstractType() {
    return Type(TypeValue::VOID);
}
bool NullAbstractNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_NullAbstractNode;
}
// class NullAbstractNode

std::shared_ptr<NullDirectAbstractNode> DirectAbstractDeclaratorNode::NULL_VAL = std::make_shared<NullDirectAbstractNode>();

DirectAbstractDeclaratorNode::DirectAbstractDeclaratorNode(ASTNodeKind nodeKind): DeclNode(AST_DirectAbstractDeclaratorNode) {};
type_info::Type DirectAbstractDeclaratorNode::generateDirectAbstractType() {
    // TODO: handle complex types; not need ATM
    return Type(TypeValue::VOID);
}

bool DirectAbstractDeclaratorNode::classof(const ASTNode * node) {
    return node->getNodeKind() >=
                            ASTNodeKind::AST_DirectAbstractDeclaratorNode &&
           node->getNodeKind() <= ASTNodeKind::AST_NullDirectAbstractNode;
}
// class DirectAbstractDeclaratorNode


NullDirectAbstractNode::NullDirectAbstractNode() : DirectAbstractDeclaratorNode(AST_DirectAbstractDeclaratorNode) {}
type_info::Type NullDirectAbstractNode::generateDirectAbstractType() {
    return Type(TypeValue::VOID);
}
bool NullDirectAbstractNode::isNull() {
    return true;
}
void NullDirectAbstractNode::unparse(int indent) {}
bool NullDirectAbstractNode::classof(const ASTNode * node) {
    return node->getNodeKind() <= ASTNodeKind::AST_NullDirectAbstractNode;
}
// end DeclNode subclasses
