//#include "symbol_table/symbol_table.hpp"

using namespace symbol_table;

template <class Symbol>
SymbolTable<Symbol>::SymbolTable() {
   static_assert(std::is_base_of<Symb, Symbol>::value, "Object is not derived from Symb");
   blockSymbolStack = std::stack<std::unordered_map<std::string, Symbol>>();
}

template <class Symbol>
void SymbolTable<Symbol>::openScope() {
  blockSymbolStack.push(std::unordered_map<std::string, Symbol>());
}

template <class Symbol>
void SymbolTable<Symbol>::closeScope() {
  if (blockSymbolStack.empty()) {
     throw EmptySTException();
  }
  else {
     blockSymbolStack.pop();
  }
}

template <class Symbol>
void SymbolTable<Symbol>::insert(const Symbol sym) {
  if (blockSymbolStack.empty()) {
     throw EmptySTException();
  }
  else {
     auto& currentScope = blockSymbolStack.top();
     if (currentScope.count(sym.get_name()) >= 1) {
        throw DuplicateException(sym.get_name());
     }
     else {
        currentScope[sym.get_name()] = sym;
     }
  }
}

template <class Symbol>
Symbol& SymbolTable<Symbol>::localLookup(const std::string name) {
  if (blockSymbolStack.empty()) {
     throw FailedLookupException(name);
  }
  else {
     auto& localScope = blockSymbolStack.top();
     if (localScope.count(name) >= 1) {
        return localScope.at(name);
     }
     else {
        throw FailedLookupException(name);
     }
  }
}

template <class Symbol>
Symbol& SymbolTable<Symbol>::globalLookup(const std::string name) {
    std::stack<std::unordered_map<std::string, Symbol>> tempStack;

    while (!blockSymbolStack.empty()) {
        auto& scope = blockSymbolStack.top();
        try {
             Symbol& result = scope.at(name);
             while(!tempStack.empty()) {
                 blockSymbolStack.push(tempStack.top());
                 tempStack.pop();
             }
             return result;
        } catch (...) {
            tempStack.push(scope);
            blockSymbolStack.pop();
            continue;
        }
    }

    // Restore stack before throwing an error when lookup fails
    while(!tempStack.empty()) {
        blockSymbolStack.push(tempStack.top());
        tempStack.pop();
    }
    throw FailedLookupException(name);
}

template <class Symbol>
SymbolTable<Symbol>::operator std::string() const {
    std::string representation = "";

    std::stack<std::unordered_map<std::string, Symbol>> tempStack(blockSymbolStack);

    while (!tempStack.empty()) {
        auto scope = tempStack.top();
        std::string join = "{ ";

        for (auto const& symbol: scope) {
            join += std::string(symbol.second);
            join += ", ";
        }

         // Remove last comma
         join.pop_back();
         join.pop_back();

         join += " }\n";
         representation += join;
         tempStack.pop();
    }

    if (representation.empty()) {
        return "--empty table--";
    }
    else {
        representation.pop_back(); // Remove last newline
        return representation;
    }

}

template <class Symbol>
void SymbolTable<Symbol>::dump(std::ostream &ps) {
    ps << std::string(*this) << "\n";
}

template <class Symbol>
int SymbolTable<Symbol>::scopeDepth() {
    return blockSymbolStack.size();
}
